/**************************************************************
 * Send any comments or questions to: OTSO-Bug@tel.vtt.fi
 *
 * Name: /home/users/otso/official/otso/include/incfirst/SCCS/s.OTSO.hxx
 * Vers: 5.2    Time: 92/08/03, 16:57:07
 **************************************************************/

/***************************************************************
* Copyright (c) 1992      Technical Research Centre of Finland (VTT)
*
* Permission to use, copy, modify, and distribute this software and its
* documentation for any purpose and without fee is hereby granted, provided
* that this notice and the reference to this notice appearing in each software
* module be retained unaltered, and that the name of any contributors shall not
* be used in advertising or publicity pertaining to distribution of the software
* without specific written prior permission.  No contributor makes any
* representations about the suitability of this software for any purpose.
* It is provided "as is" without any express or limited warranty.
*
*			NO WARRANTY
*
* ALL CONTRIBUTORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
* INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS.  IN NO
* EVENT SHALL ANY CONTRIBUTOR BE LIABLE FOR ANY SPECIAL, PUNITIVE, INDIRECT OR
* CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
* DATA, OR PROFITS, WHETHER IN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
* TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH, THE USE OR PERFORMANCE
* OF THIS SOFTWARE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THIS
* SOFTWARE IS WITH YOU.  SHOULD THIS SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE
* COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
*
* As used above, "contributor" includes, but is not limited to :
*        The Technical Research Centre of Finland
***************************************************************/


#ifndef OTSO_HXX
#define OTSO_HXX


#if 0
/*
 * This is a general OTSO header file. Include this into every .cxx
 * module dealing with OTSO!
 *
 * OTSO developers might do the following (in OTSO's kernel .cxx:s):
 *
 * 1. #define as 1 only the names of headers you want to include
 * 2. #define OTSO_SELECT_INCLUDES as 1
 * 3. #include "OTSO.hxx"
 *
 * and OTSO users would probably just (in application .cxx:s):
 * 
 * 1. #include "OTSO.hxx"
 * 
 * 
 * IMPORTANT NOTE.1:
 * 	Other OTSO kernel header files don't need anymore extraneous
 * 	#defines to know if they are already included. It is handled
 * 	here.
 * 
 * IMPORTANT NOTE.2:
 * 	If new header files are added to OTSO or old headers
 * 	are deleted, modify this file accordingly !!!
 * 
 * IMPORTANT NOTE.3:
 * 	#includes in this file should be ordered properly !!!
 * 
 * 
 * -------------- start of example -----------------
 * 
 * // Example.cxx:
 * 
 * #define OBJECT_HXX		1 // I want these headers included 
 * #define FRAME_HXX		1
 * ...
 * ...
 * #define OTSO_SELECT_INCLUDES	1 // Prevent including other OTSO headers; 
 * 				  // this speeds up compilation! 
 * #include "OTSO.hxx"		  // Only this line would be enough!
 * 
 * -------------- end of example -----------------
 */
#endif  /* #if 0 */


#include "env_sel.h"
#include "envCC.h"
#include "envOS.h"
#include "envOTSO.h"



/* make this usable in #if:s here */
#ifdef OTSO_SELECT_INCLUDES
# if (OTSO_SELECT_INCLUDES != 0) && (OTSO_SELECT_INCLUDES != 1)
#  undef OTSO_SELECT_INCLUDES
#  define OTSO_SELECT_INCLUDES	1
# endif
#else
# define OTSO_SELECT_INCLUDES	0
#endif


#if (!OTSO_SELECT_INCLUDES) || (OBJECT_HXX == 1)
#include "object.hxx"
#endif

#if (!OTSO_SELECT_INCLUDES) || (STRING_HXX == 1)
#include "string.hxx"
#endif

#if (!OTSO_SELECT_INCLUDES) || (BYTES_HXX == 1)
#include "bytes.hxx"
#endif

#if (!OTSO_SELECT_INCLUDES) || (MULTI_HXX == 1)
#include "multi.hxx"
#endif

#if (!OTSO_SELECT_INCLUDES) || (TYPE_HXX == 1)
#include "type.hxx"
#endif

#if (!OTSO_SELECT_INCLUDES) || (TIME_HXX == 1)
#include "time.hxx"
#endif


#if (SIMULATING || STATISTICS)
  // *MJS* changes
# if (!OTSO_SELECT_INCLUDES) || (STATOBJ_HXX == 1)
#  include "tree.hxx"		/*?*/
#  include "simres.hxx"		/*?*/
#  include "statobj.hxx"
# endif
# if (!OTSO_SELECT_INCLUDES) || (SAMPLED_HXX == 1)
#  include "sampled.hxx"
# endif
#endif	/* STATISTICS GATHERING */


#if (!OTSO_SELECT_INCLUDES) || (INT_HXX == 1)
#include "int.hxx"
#endif

#if (!OTSO_SELECT_INCLUDES) || (NAMEDOBJ_HXX == 1)
#include "namedobj.hxx"
#endif

#if (!OTSO_SELECT_INCLUDES) || (GROUP_HXX == 1)
#include "group.hxx"
#endif

#if (!OTSO_SELECT_INCLUDES) || (RUNNER_HXX == 1)
#include "runner.hxx"
#endif


#if (!OTSO_SELECT_INCLUDES) || (DEVICE_HXX == 1)
#include "device.hxx"
#endif

#if (!OTSO_SELECT_INCLUDES) || (DEVXFILE_HXX == 1)
#include "devxfile.hxx"
#endif

#if BSD	/*?*/ 
# if (!OTSO_SELECT_INCLUDES) || (DEVSOCK_HXX == 1)
#  include "devsock.hxx"
# endif
#endif

#if (!OTSO_SELECT_INCLUDES) || (CHANNEL_HXX == 1)
#include "channel.hxx"
#endif

#if (!OTSO_SELECT_INCLUDES) || (AGENT_HXX == 1)
#include "agent.hxx"
#endif

#if (!OTSO_SELECT_INCLUDES) || (MESSAGE_HXX == 1)
#include "message.hxx"
#endif

#if (!OTSO_SELECT_INCLUDES) || (PROCESS_HXX == 1)
#include "process.hxx"
#endif


#if (!OTSO_SELECT_INCLUDES) || (VOIDGRP_HXX == 1)
#include "voidgrp.hxx"
#endif

#if (!OTSO_SELECT_INCLUDES) || (DYNDIR_HXX == 1)
#include "dyndir.hxx"
#endif


#if (!OTSO_SELECT_INCLUDES) || (FIFO_HXX == 1)
#include "fifo.hxx"
#endif

/**************************************************
* #if (!OTSO_SELECT_INCLUDES) || (LIFO_HXX == 1)
* #include "lifo.hxx"
* #endif
**************************************************/

#if (!OTSO_SELECT_INCLUDES) || (RING_HXX == 1)
#include "ring.hxx"
#endif

#if (!OTSO_SELECT_INCLUDES) || (PRIORVEC_HXX == 1)
#include "priorvec.hxx"
#endif


#if (!OTSO_SELECT_INCLUDES) || (INTA_HXX == 1)
#include "inta.hxx"
#endif


#if (!OTSO_SELECT_INCLUDES) || (FRAME_HXX == 1)
#include "frame.typ"
#include "frame.ext"
#include "frame.hxx"
#endif


#if (!OTSO_SELECT_INCLUDES) || (STR2VEC_HXX == 1)
#include "str2vec.hxx"
#endif


#if DEBUGGING_FREE_STORE_MEMORY
# if (!OTSO_SELECT_INCLUDES) || (MEMDBG_HXX == 1)
#  include "memdbgSP.hxx"
#  include "memdbg.hxx"
# endif
#endif /* DEBUGGING_FREE_STORE_MEMORY */

#endif    /* OTSO_HXX */
